/* dhcpd.c
 *
 * udhcp Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <network.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>

#include "ecos.h"
#include "socket.h"
#include "options.h"
#include "files.h"
#include "packet.h"
#include "serverpacket.h"
#include <net/dhcpelios/debug.h>
#include <net/dhcpelios/dhcpd.h>
#include <net/dhcpelios/leases.h>

// Global Variables
struct dhcpOfferedAddr *leases=NULL;						// Leases table
struct server_config_t server_config;						// Configuration parameters
char dhcp_running=0;								// Boolean (running loop)
char dhcp_started=0;								// Boolean (started or not)
int server_socket=-1;								// Server listening socket

// Embedded Part
static cyg_uint8 dhcpd_stack[CYGNUM_ELIOSDHCP_THREAD_STACK_SIZE]; 			// Stack for DHCP thread
static cyg_handle_t dhcpd_thread;							// Handle on DHCP thread
static cyg_thread  dhcpd_thread_object;							// DHCP Thread Object

//-----------------------------------------------------------------------------------------
// Exit and cleanup
// @Param  : retval Status code to return
// @Return : Status code
//-----------------------------------------------------------------------------------------
static int elios_dhcpd_exit(const int retval)
{
 dhcp_running=0;
 DEBUG("elios_dhcpd_exit","Called");
 if (server_socket>0)
 {
  close(server_socket);
  LOG(LOG_INFO,"elios_dhcpd_exit","Closing server socket %d",server_socket);
  server_socket=-1;
 }
 LOG(LOG_INFO,"elios_dhcpd_exit","Freeing leases");
 if (leases) free(leases);
 LOG(LOG_INFO,"elios_dhcpd_exit","Freeing options");
 if (server_config.options) FreeOptionsList(server_config.options);
 LOG(LOG_INFO,"elios_dhcpd_exit","Freeing strdup allocations");
 if (server_config.interface) free(server_config.interface);
 if (server_config.lease_file) free(server_config.lease_file);
 if (server_config.sname) free(server_config.sname);
 if (server_config.boot_file) free(server_config.boot_file);
 LOG(LOG_INFO,"elios_dhcpd_exit","Ending");
 dhcp_started=0;
 return(retval);
}

//-----------------------------------------------------------------------------------------
// DHCP server thread entry point
// @ returns : Nothing
//-----------------------------------------------------------------------------------------
static void elios_dhcpd_init(cyg_addrword_t arg)
{
 fd_set rfds;						// Socket descriptors to scan
 struct timeval tv;					// Timeout on sockets events
 int bytes=0;						// Size of received data
 int retval=0;						// Returned value
 struct dhcpMessage packet;				// Reveived packet
 unsigned char *state=NULL;				// Type of DHCP message
 unsigned char *server_id=NULL;				// IP address of server
 unsigned char *requested=NULL;				// IP address requested by client
 u_int32_t server_id_align=0;				// IP address of server (aligned)
 u_int32_t requested_align=0;				// IP address requested by client (aligned)
 unsigned long timeout_end=0;				// Lease Timeout
 unsigned long thetime=0;				// Current time
 struct option_set *option=NULL;			// Options DHCP
 struct dhcpOfferedAddr *lease=NULL;			// Choosed Lease
 char arp[MACSTRING];					// Human form of our ARP address
#ifdef LOGACTIVE
 char ip[IPSTRING];					// IP address conversion
#endif

 DEBUG("elios_dhcpd_init","Called");
 // Set the running flag (loop)
 dhcp_running=1;
 // Empty configuration structure
 memset(&server_config, 0, sizeof(struct server_config_t));

 // Read config file or set default values
 read_config(DHCPD_CONF_FILE);

 LOG(LOG_INFO,"elios_dhcpd_init","IP pool start: '%s'",DHCPNumber2String(server_config.start,ip));
 LOG(LOG_INFO,"elios_dhcpd_init","IP pool end :'%s'",DHCPNumber2String(server_config.end,ip));
 LOG(LOG_INFO,"elios_dhcpd_init","Max leases :%lu",server_config.max_leases);
 LOG(LOG_INFO,"elios_dhcpd_init","Min lease :%lu",server_config.min_lease);
#ifdef CYGPKG_ELIOSDHCP_LEASE_FILE
 LOG(LOG_INFO,"elios_dhcpd_init","Lease file :'%s'",server_config.lease_file);
#endif
 LOG(LOG_INFO,"elios_dhcpd_init","Auto time  :%lu",server_config.auto_time);
 LOG(LOG_INFO,"elios_dhcpd_init","Decline time  :%lu",server_config.decline_time);
 LOG(LOG_INFO,"elios_dhcpd_init","Offer time  :%lu",server_config.offer_time);
 LOG(LOG_INFO,"elios_dhcpd_init","Conflict time  :%lu",server_config.conflict_time);
 LOG(LOG_INFO,"elios_dhcpd_init","Remaining :%s",server_config.remaining?"y":"n");
 LOG(LOG_INFO,"elios_dhcpd_init","Working Interface  :'%s'",server_config.interface);
 LOG(LOG_INFO,"elios_dhcpd_init","Sname  :'%s'",server_config.sname);
 LOG(LOG_INFO,"elios_dhcpd_init","Boot file  :'%s'",server_config.boot_file);

 // Look for LEASE_TIME in server options
 if ((option = find_option(server_config.options, DHCP_LEASE_TIME)))
 {
  memcpy(&server_config.lease, option->data+OPT_DATA, sizeof(server_config.lease));
  server_config.lease = ntohl(server_config.lease); // Convert in host order
 }
 else server_config.lease = LEASE_TIME; // default value
 LOG(LOG_INFO,"elios_dhcpd_init","Lease time:%lu/%d",server_config.lease,LEASE_TIME);

 // Lease table allocation
 leases = malloc(sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
 if (!leases)
 {
  LOG(LOG_CRIT,"elios_dhcpd_init","Could not allocate leases table of size %lu*%u=%lu bytes: '%s'",server_config.max_leases,sizeof(struct dhcpOfferedAddr),sizeof(struct dhcpOfferedAddr)*server_config.max_leases,strerror(errno));
  elios_dhcpd_exit(-1);
 }

 // Reset leases
 memset(leases, 0, sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
#ifdef CYGPKG_ELIOSDHCP_LEASE_FILE
 read_leases(server_config.lease_file);		// Read leases file if it exists
#endif

 // Read server network configuration
 if ( (read_interface(server_config.interface,&server_config.ifindex,&server_config.server,server_config.arp)<0))
 {
  LOG(LOG_CRIT,"elios_dhcpd_init","Could not obtain network interface %s address :'%s'",server_config.interface,strerror(errno));
  elios_dhcpd_exit(-2);
 }
#ifdef LOGACTIVE
 LOG(LOG_INFO,"elios_dhcpd_init","Server :'%s'",DHCPNumber2String(server_config.server,ip));
 DHCPether_print(server_config.arp,arp,MACSTRING);
 LOG(LOG_INFO,"elios_dhcpd_init","Server interface:'%s' No[%d] ARP:'%s'",server_config.interface,server_config.ifindex,arp);
#endif
 // Time left before writting leases
 timeout_end = time(0) + server_config.auto_time;
 while(dhcp_running)
 { // loop until universe collapses : until we got a listening socket
  thetime=time(0);				// Current time
  DEBUG("elios_dhcpd_init","Timeout end=%lu time=%lu,auto_time=%lu",timeout_end,thetime,server_config.auto_time);
  // Server socket creation
  if (server_socket < 0)
  {
   // We will listen broadcast packets to SERVER_PORT
   if ((server_socket = listen_socket(INADDR_ANY, SERVER_PORT, server_config.interface)) < 0)
   {
     LOG(LOG_CRIT,"elios_dhcpd_init","Couldn't create server listening socket on port %d on interface '%s':'%s'",SERVER_PORT,server_config.interface, strerror(errno));
     elios_dhcpd_exit(-3);
   }
   else LOG(LOG_INFO,"elios_dhcpd_init","Server socket no %d created on port %d on interface '%s'",server_socket,SERVER_PORT,server_config.interface);
  }
  // Waiting read events on server_socket
  FD_ZERO(&rfds);
  FD_SET(server_socket, &rfds);
  if (server_config.auto_time) // Non blocking mode for writting leases each server_config.auto_time
  {
   tv.tv_sec = timeout_end - time(0); // Timeout for writting leases
   tv.tv_usec = 0;
  }
  if (!server_config.auto_time || tv.tv_sec > 0)
  {
   // Wait for a packet ...
   DEBUG("elios_dhcpd_init","Before select on socket %d",server_socket);
   // Blocking call
   retval = select(server_socket + 1, &rfds, NULL, NULL,server_config.auto_time ? &tv : NULL);
   DEBUG("elios_dhcpd_init","After select on socket %d",server_socket);
  }
  else retval = 0; // If we already timed out, fall through
  if (retval == 0)
  {
   // timeout, we write lease files
#ifdef CYGPKG_ELIOSDHCP_LEASE_FILE
   write_leases();
#endif
   // Do it again
   timeout_end = time(0) + server_config.auto_time;
   continue;
  }
  else if (retval < 0 && errno != EINTR)
  {
   LOG(LOG_ERR,"elios_dhcpd_init","Error on select :'%s'",strerror(errno));
   continue;
  }

  // Read received packet
  if ((bytes = get_packet(&packet, server_socket)) < 0)
  { // this waits for a packet - idle
   if (bytes == -1 && errno != EINTR)
   {
    LOG(LOG_ERR,"elios_dhcpd_init","Error on read, %s, closing socket no %d", strerror(errno),server_socket);
    close(server_socket);
    server_socket = -1;
   }
   continue;
  }
  else LOG(LOG_INFO,"elios_dhcpd_init","Received %d bytes on socket %d",bytes,server_socket);

  if ((state = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL)
  {
   LOG(LOG_ERR,"elios_dhcpd_init","Couldn't get message type from packet");
   continue;
  }

  DHCPether_print(packet.chaddr,arp,MACSTRING);
  // Find a lease
  LOG(LOG_INFO,"elios_dhcpd_init","State : %d Wanted lease :'%s'",state[0],arp);
  lease = find_lease_by_chaddr(packet.chaddr);
  if (lease)
  {
   LOG(LOG_INFO,"elios_dhcpd_init","Lease choosed: '%s' for MAC '%s'",DHCPNumber2String(lease->yiaddr,ip),arp);
  }
  else LOG(LOG_INFO,"elios_dhcpd_init","No lease for '%s' in our table",arp);
#ifndef CYGPKG_ELIOSDHCP_MANUAL_CONF
  // Compute automatic values
  read_network_config();
#endif
  switch (state[0])
  {
   case DHCPDISCOVER:
    LOG(LOG_INFO,"elios_dhcpd_init","Received DISCOVER");
    if (sendOffer(&packet) < 0)
    {
     LOG(LOG_ERR,"elios_dhcpd_init","Send OFFER failed");
    }
    break;
   case DHCPREQUEST:
    LOG(LOG_INFO,"elios_dhcpd_init","Received REQUEST");
    requested = get_option(&packet, DHCP_REQUESTED_IP);
    server_id = get_option(&packet, DHCP_SERVER_ID);
    if (requested) memcpy(&requested_align, requested, sizeof(requested_align));
    if (server_id) memcpy(&server_id_align, server_id, sizeof(server_id_align));
    if (lease)
    {
     if (server_id) // Client got a server id
     {
      // SELECTING State
      LOG(LOG_INFO,"elios_dhcpd_init","Server_id = %08x '%s'",ntohl(server_id_align),DHCPNumber2String(server_id_align,ip));
      LOG(LOG_INFO,"elios_dhcpd_init","IP requested = %08x '%s'", ntohl(requested_align),DHCPNumber2String(requested_align,ip));
      LOG(LOG_INFO,"elios_dhcpd_init","Lease yiaddr = %08x '%s'", ntohl(lease->yiaddr),DHCPNumber2String(lease->yiaddr,ip));
      if (server_id_align == server_config.server && requested && requested_align == lease->yiaddr)
      {
       LOG(LOG_INFO,"elios_dhcpd_init","Aknowledging for %s...",DHCPNumber2String(lease->yiaddr,ip));
       // This is for us !
       sendACK(&packet, lease->yiaddr);
      }
      else LOG(LOG_INFO,"elios_dhcpd_init","This is for another DHCP server (%s/%s) or (%s!=%s)...",DHCPNumber2String(server_id_align,ip),DHCPNumber2String(server_config.server,ip),DHCPNumber2String(requested_align,ip),DHCPNumber2String(lease->yiaddr,ip));
     }
     else
     {
      LOG(LOG_INFO,"elios_dhcpd_init","No server id in request");
      if (requested)
      {
       LOG(LOG_INFO,"elios_dhcpd_init","Requested %08x '%s'",ntohl(requested_align),DHCPNumber2String(requested_align,ip));
       // INIT-REBOOT State
       if (lease->yiaddr == requested_align)
        sendACK(&packet, lease->yiaddr);
       else sendNAK(&packet);
      }
      else
      {
       LOG(LOG_INFO,"elios_dhcpd_init","No server id, No IP requested :ciaddr is '%s'",DHCPNumber2String(packet.ciaddr,ip));
       // RENEWING or REBINDING State
       if (lease->yiaddr == packet.ciaddr)
        sendACK(&packet, lease->yiaddr);
       else
        sendNAK(&packet);
      }
     }
     // what to do if we have no record of the client
    }
    else // No lease found with packet.chaddr
    {
     LOG(LOG_INFO,"elios_dhcpd_init","No lease for '%s'",arp);
     if (server_id)
     {
      // SELECTING State
       LOG(LOG_INFO,"elios_dhcpd_init","SELECTING state");
     }
     else if (requested)
     {
      // INIT-REBOOT State
       LOG(LOG_INFO,"elios_dhcpd_init","INIT-REBOOT state . Old IP:'%s'",DHCPNumber2String(requested_align,ip));
      if ((lease = find_lease_by_yiaddr(requested_align)))
      {
       if (lease_expired(lease))
       {
        LOG(LOG_INFO,"elios_dhcpd_init","Lease for %s has expired",DHCPNumber2String(requested_align,ip));
        // probably best if we drop this lease
        memset(lease->chaddr, 0, TABLELEASESIZE);
        // make some contention for this address
       }
       else sendNAK(&packet);
      }
      else if (requested_align < server_config.start ||requested_align > server_config.end)
      {
       LOG(LOG_INFO,"elios_dhcpd_init","Requested address %s is not in pool range",DHCPNumber2String(requested_align,ip));
       sendNAK(&packet);
      } // else remain silent
     }
     else
     {
      LOG(LOG_INFO,"elios_dhcpd_init","We do nothing");
      // RENEWING or REBINDING State
     }
    }
    break;
   case DHCPDECLINE:
    LOG(LOG_INFO,"elios_dhcpd_init","Received DECLINE");
    if (lease)
    {
     // probably best if we drop this lease
     memset(lease->chaddr, 0, TABLELEASESIZE);
     lease->expires = time(0) + server_config.decline_time;
    }
    break;
   case DHCPRELEASE:
    LOG(LOG_INFO,"elios_dhcpd_init","Received RELEASE");
    // Renewing time
    if (lease) lease->expires = time(0); // expires now !
    break;
   case DHCPINFORM:
    LOG(LOG_INFO,"elios_dhcpd_init","Received INFORM");
    send_inform(&packet);
    break;
   default:
    LOG(LOG_WARNING,"elios_dhcpd_init","Unsupported DHCP message (%02x) -- ignoring", state[0]);
  }
 } // Fin du while dhcp_running
 elios_dhcpd_exit(0);
}


//---------------------------------------------------------------------------------------
// Thread entry point
// Start DHCP server
// @Return -1 on error and 0 if OK
//-----------------------------------------------------------------------------------------
__externC int dhcpd_startup(void)
{
 DEBUG("dhcpd_startup","Called");
 if (dhcp_started==1)
 {
  LOG(LOG_ERR,"dhcpd_startup","DHCP has already been startup");
  return(-1);
 }
 if (dhcp_running==0)
 {
  cyg_thread_create(CYGNUM_ELIOSDHCP_THREAD_PRIORITY,
  elios_dhcpd_init,
  0,
  "UDHCP server",
   dhcpd_stack,
   sizeof(dhcpd_stack),
   &dhcpd_thread,
   &dhcpd_thread_object);
   LOG(LOG_INFO,"dhcpd_startup","Starting DHCP thread");
   cyg_thread_resume(dhcpd_thread);
  dhcp_started=1;
  return(0);
 }
 else
 {
  LOG(LOG_ERR,"dhcpd_startup","DHCP is already running");
  return(-1);
 }
}

//-----------------------------------------------------------------------------------------
// Stop DHCP Server
// @Return -1 on error and 0 if OK
//-----------------------------------------------------------------------------------------
__externC int dhcpd_stop(void)
{
 int s=0;
 struct sockaddr_in server;

 DEBUG("dhcpd_stop","Stopping DHCP thread");
 dhcp_running=0;
 s=socket(AF_INET, SOCK_DGRAM, 0);
 if (s<0)
 {
  LOG(LOG_ERR,"Cannot create UDP socket:'%s'",strerror(errno));
  return(-1);
 }
 else
 {
  memset(&server, 0, sizeof(server));
  server.sin_family = AF_INET;
  server.sin_len = sizeof(server);
  server.sin_port = htons(CYGNUM_ELIOSDHCP_PORT);
  server.sin_addr.s_addr=inet_addr("127.0.0.1");
  sendto(s,"END",3,0,(struct sockaddr *)&server,sizeof(server));
  close(s);
 }
 return(0);
}

//-----------------------------------------------------------------------------------------
// Suspending DHCP Server
//-----------------------------------------------------------------------------------------
__externC void dhcpd_suspend(void)
{
 DEBUG("dhcpd_suspend","Suspending DHCP thread");
 cyg_thread_suspend(dhcpd_thread);
}

//-----------------------------------------------------------------------------------------
// Stopping DHCP Server
//-----------------------------------------------------------------------------------------
__externC void dhcpd_exit(void)
{
 DEBUG("dhcpd_exit","Exitting DHCP thread");
 elios_dhcpd_exit(0);
}
