/*
 * files.c -- DHCP server file manipulation
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * udhcp Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include "ecos.h"

#include "files.h"
#include "options.h"
#include "socket.h"
#include <net/dhcpelios/debug.h>
#include <net/dhcpelios/leases.h>
#include <net/dhcpelios/dhcpd.h>

//-----------------------------------------------------------------------------------------
// on these functions, make sure you datatype matches
// Read IP adress
// Return 1 on success, 0 otherwise
//-----------------------------------------------------------------------------------------
static int read_ip(char *line, void *arg)
{
	struct in_addr *addr = arg;
	struct hostent *host=NULL;
	int retval = 1;

	DEBUG("read_ip","Called");
	if (!inet_aton(line, addr))
	{
		if ((host = gethostbyname(line)))
			addr->s_addr = *((unsigned long *) host->h_addr_list[0]);
		else retval = 0;
	}
	return (retval);
}

//-----------------------------------------------------------------------------------------
// on these functions, make sure you datatype matches
// Read string
// Return 1 on success, 0 otherwise
//-----------------------------------------------------------------------------------------
static int read_str(char *line, void *arg)
{
	char **dest = arg;
	DEBUG("read_str","Called");
	DEBUG("read_str","Line: '%s'",line);
	DEBUG("read_str","%p '%s'",dest,*dest);
	if (*dest) free(*dest);
	*dest = strdup(line);  // Beware of memory leaks
	return (1);
}

//-----------------------------------------------------------------------------------------
// on these functions, make sure you datatype matches
// Read unsigned int
// Return 1 on success, 0 otherwise
//-----------------------------------------------------------------------------------------
static int read_u32(char *line, void *arg)
{
	u_int32_t *dest = arg;
	char *endptr=NULL;
	DEBUG("read_u32","Called");

	*dest = strtoul(line, &endptr, 0);
	return endptr[0] == '\0';
}


//-----------------------------------------------------------------------------------------
// on these functions, make sure you datatype matches
// Read Boolean (yes/no)
// Return 1 on success, 0 otherwise
//-----------------------------------------------------------------------------------------
static int read_yn(char *line, void *arg)
{
	char *dest = arg;
	int retval = 1;
	DEBUG("read_yn","Called");

	if (!strcasecmp("yes", line))
		*dest = 1;
	else if (!strcasecmp("no", line))
		*dest = 0;
	else retval = 0;

	return (retval);
}

//-----------------------------------------------------------------------------------------
// read a dhcp option and add it to opt_list
//-----------------------------------------------------------------------------------------
static int read_opt(char *line, void *arg)
{
	struct option_set **opt_list = arg;
	char *opt=NULL;
	char *val=NULL;
	char *endptr=NULL;
	struct dhcp_option *option = NULL;
	int retval = 0;
	int length = 0;
	char buffer[255];
	char buffer2[255];
	u_int16_t result_u16=0;
	u_int32_t result_u32=0;
	int i=0;

	DEBUG("read_opt","Called with '%s'",line);
	if (!(opt = strtok_r(line, " \t=",(char**)&buffer2)))
	{
	 LOG(LOG_ERR,"read_opt","Option '%s' not recognised !",line);
	 return 0;
	}
	LOG(LOG_INFO,"read_opt","strtok passed");

	LOG(LOG_INFO,"read_opt","opt='%s'",opt);
	for (i = 0; options[i].code; i++)
		if (!strcmp(options[i].name, opt))
			option = &(options[i]);

	if (!option) { LOG(LOG_ERR,"read_opt","Option '%s' unknonw !",opt); return (0); }

	do {
		val = strtok_r(NULL, ", \t",(char**)&buffer2);
		LOG(LOG_INFO,"read_opt","val='%s'",val);
		if (val) {
			length = option_lengths[option->flags & TYPE_MASK];
			retval = 0;
			switch (option->flags & TYPE_MASK) {
			case OPTION_IP:
				retval = read_ip(val, buffer);
				break;
			case OPTION_IP_PAIR:
				retval = read_ip(val, buffer);
				if (!(val = strtok_r(NULL, ", \t/-",(char**)buffer2))) retval = 0;
				if (retval) retval = read_ip(val, buffer + 4);
				break;
			case OPTION_STRING:
				length = strlen(val);
				if (length > 0) {
					if (length > sizeof(buffer)-1) length = sizeof(buffer)-1;
					memcpy(buffer, val, length);
					retval = 1;
				}
				break;
			case OPTION_BOOLEAN:
				retval = read_yn(val, buffer);
				break;
			case OPTION_U8:
				buffer[0] = strtoul(val, &endptr, 0);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_U16:
				result_u16 = htons(strtoul(val, &endptr, 0));
				memcpy(buffer, &result_u16, 2);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_S16:
				result_u16 = htons(strtol(val, &endptr, 0));
				memcpy(buffer, &result_u16, 2);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_U32:
				result_u32 = htonl(strtoul(val, &endptr, 0));
				memcpy(buffer, &result_u32, 4);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_S32:
				result_u32 = htonl(strtol(val, &endptr, 0));
				memcpy(buffer, &result_u32, 4);
				retval = (endptr[0] == '\0');
				break;
			default:
				break;
			}
			if (retval)
				attach_option(opt_list, option, buffer, length);
		};
	} while (val && retval && option->flags & OPTION_LIST);
	return (retval);
}

//-----------------------------------------------------------------------------------------
// Set a dhcp option and add it to opt_list
//-----------------------------------------------------------------------------------------
static int set_opt(char* opt,char* val, void *arg)
{
	struct option_set **opt_list = arg;
	char *endptr=NULL;
	struct dhcp_option *option = NULL;
	int retval = 0;
	int length = 0;
	char buffer[255];
	char buffer2[255];
	u_int16_t result_u16=0;
	u_int32_t result_u32=0;
	int i=0;

	DEBUG("set_opt","opt='%s'",opt);
	for (i = 0; options[i].code; i++)
		if (!strcmp(options[i].name, opt))
			option = &(options[i]);

	if (!option) { LOG(LOG_ERR,"set_opt","Option '%s' not recognised !",opt); return (0); }

		if (val) {
			LOG(LOG_INFO,"set_opt","val='%s'",val);
			length = option_lengths[option->flags & TYPE_MASK];
			retval = 0;
			switch (option->flags & TYPE_MASK) {
			case OPTION_IP:
				retval = read_ip(val, buffer);
				break;
			case OPTION_IP_PAIR:
				retval = read_ip(val, buffer);
				if (!(val = strtok_r(NULL, ", \t/-",(char**)buffer2))) retval = 0;
				if (retval) retval = read_ip(val, buffer + 4);
				break;
			case OPTION_STRING:
				length = strlen(val);
				if (length > 0) {
					if (length > sizeof(buffer)-1) length = sizeof(buffer)-1;
					memcpy(buffer, val, length);
					retval = 1;
				}
				break;
			case OPTION_BOOLEAN:
				retval = read_yn(val, buffer);
				break;
			case OPTION_U8:
				buffer[0] = strtoul(val, &endptr, 0);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_U16:
				result_u16 = htons(strtoul(val, &endptr, 0));
				memcpy(buffer, &result_u16, 2);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_S16:
				result_u16 = htons(strtol(val, &endptr, 0));
				memcpy(buffer, &result_u16, 2);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_U32:
				result_u32 = htonl(strtoul(val, &endptr, 0));
				memcpy(buffer, &result_u32, 4);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_S32:
				result_u32 = htonl(strtol(val, &endptr, 0));
				memcpy(buffer, &result_u32, 4);
				retval = (endptr[0] == '\0');
				break;
			default:
				break;
			}
			if (retval)
				attach_option(opt_list, option, buffer, length);
		};
	return (retval);
}

#ifndef CYGNUM_ELIOSDHCP_INTERFACE
#define CYGNUM_ELIOSDHCP_INTERFACE "eth1"
#endif

#ifndef CYGNUM_ELIOSDHCP_START_ADDR_IP
#define CYGNUM_ELIOSDHCP_START_ADDR_IP "10.0.0.2"
#endif

#ifndef CYGNUM_ELIOSDHCP_END_ADDR_IP
#define CYGNUM_ELIOSDHCP_END_ADDR_IP "10.0.0.253"
#endif

#ifndef CYGNUM_ELIOSDHCP_LEASE_FILENAME
#define CYGNUM_ELIOSDHCP_LEASE_FILENAME ""
#endif

#ifndef CYGNUM_ELIOSDHCP_LEASE_AUTO_TIME
#define CYGNUM_ELIOSDHCP_LEASE_AUTO_TIME "120"
#endif

#define stringification(x) #x

//-----------------------------------------------------------------------------------------
// Valeurs par dfaut
//-----------------------------------------------------------------------------------------
static struct config_keyword keywords[] = {
#ifdef CYGPKG_ELIOSDHCP_MANUAL_CONF
	{"start",	read_ip,  &(server_config.start),	CYGNUM_ELIOSDHCP_START_ADDR_IP}, // Pool start
	{"end",		read_ip,  &(server_config.end),		CYGNUM_ELIOSDHCP_END_ADDR_IP},
 // Pool end
	{"mask",	read_ip,  &(server_config.mask),	CYGNUM_ELIOSDHCP_ADDR_MASK}, // Netmask
#endif
	{"interface",	read_str, &(server_config.interface),	CYGNUM_ELIOSDHCP_INTERFACE},
   // Answering interface
	{"option",	read_opt, &(server_config.options),	""},
	{"opt",		read_opt, &(server_config.options),	""},
	{"max_leases",	read_u32, &(server_config.max_leases),	"254"},			// Maximum number of leases
	{"remaining",	read_yn,  &(server_config.remaining),	"yes"},			// Remaining
	{"auto_time",	read_u32, &(server_config.auto_time),	CYGNUM_ELIOSDHCP_LEASE_AUTO_TIME},			// Interval between leases write to disk !
	{"decline_time",read_u32, &(server_config.decline_time),"3600"},		// Decline time
	{"conflict_time",read_u32,&(server_config.conflict_time),"1800"},		// Conflict time
	{"offer_time",	read_u32, &(server_config.offer_time),	"100"},			// Offer time
	{"min_lease",	read_u32, &(server_config.min_lease),	"60"},			// Minimum lease time
#ifdef CYGNUM_ELIOSDHCP_LEASE_FILENAME
	{"lease_file",	read_str, &(server_config.lease_file),	CYGNUM_ELIOSDHCP_LEASE_FILENAME}, 	// Name of lease file
#endif
	{"siaddr",	read_ip,  &(server_config.siaddr),	"0.0.0.0"},		// Server IP (don't change here)
	{"sname",	read_str, &(server_config.sname),	""},			// Server name
	{"boot_file",	read_str, &(server_config.boot_file),	""},			// Boot file name
	{"",		NULL, 	  NULL,				""}			// END
};

//-----------------------------------------------------------------------------------------
// Read configuration File
// @Return 1 on success , 0 otherwize
//-----------------------------------------------------------------------------------------
char read_network_config(void)
{
 char serv[IPSTRING];
 char mask[IPSTRING];
#ifndef CYGPKG_ELIOSDHCP_MANUAL_CONF
  u_int32_t newmask;			// Automatic netmask
#endif

 // We place DHCP server as the default gateway and default DNS and we give subnet mask
 if (read_interface(server_config.interface, &server_config.ifindex,&server_config.server, server_config.arp) < 0)
 {
  LOG(LOG_ERR,"read_network_config","Cannot obtain interface %s address :'%s'",server_config.interface,strerror(errno));
  return(0);
 }
 DHCPNumber2String(server_config.server,serv);
#ifndef CYGPKG_ELIOSDHCP_MANUAL_CONF
 // Automatic configuration :
 //  - Netmask is the interface one. (server_config.interface)
 //  - IP pool start address is interface adress+1
 //  - IP pool end address is interface adress+server_config.max_leases;
 if (read_interface_mask(server_config.interface,&newmask)<0)
   LOG(LOG_ERR,"read_network_config","Cannot obtain interface %s netmask : '%s'",server_config.interface,strerror(errno));
 else
 {
  LOG(LOG_INFO,"read_network_config","Automatic Netmask is '%s'",DHCPNumber2String(newmask,mask));
  server_config.mask=newmask;
 }
 server_config.start=ntohl(server_config.server)+1;
 server_config.end=ntohl(server_config.start)+server_config.max_leases;
#endif
 DHCPNumber2String(server_config.mask,mask);
 LOG(LOG_INFO,"read_network_config","Setting gw and dns to '%s' and netmask to '%s'", serv,mask);
 set_opt("subnet",mask,&(server_config.options));
 set_opt("dns",serv,&(server_config.options));
 set_opt("router",serv,&(server_config.options));
 return (1);
}


//-----------------------------------------------------------------------------------------
// Read configuration File
// @param file : filename
// @Return 1
//-----------------------------------------------------------------------------------------
char read_config(char *file)
{
	FILE *in=NULL;
	char buffer[80];
	char orig[80];
	char *token=NULL;
	char *line=NULL;
	int i=0;

	DEBUG("read_config","Called with '%s'",file);
	// Default Values
	for (i = 0; strlen(keywords[i].keyword); i++)
		if (strlen(keywords[i].def))
			keywords[i].handler(keywords[i].def, keywords[i].var);

	if (!(in = fopen(file, "r"))) {
		LOG(LOG_ERR,"read_config","Unable to open DHCP configuration file '%s' :'%s'", file,strerror(errno));

		read_network_config();
		return (1);
	}

	// Parse config file
	while (fgets(buffer,sizeof(buffer), in)) {
		if (strchr(buffer, '\n')) *(strchr(buffer, '\n')) = '\0';
		strncpy(orig, buffer,sizeof(buffer));
		if (strchr(buffer, '#')) *(strchr(buffer, '#')) = '\0';
		token = buffer + strspn(buffer, " \t");
		if (*token == '\0') continue;
		line = token + strcspn(token, " \t=");
		if (*line == '\0') continue;
		*line = '\0';
		line++;

		// eat leading whitespace
		line = line + strspn(line, " \t=");
		// eat trailing whitespace
		for (i = strlen(line); i > 0 && isspace(line[i - 1]); i--);
		line[i] = '\0';

		for (i = 0; strlen(keywords[i].keyword); i++)
			if (!strcasecmp(token, keywords[i].keyword))
				if (!keywords[i].handler(line, keywords[i].var)) {
					LOG(LOG_ERR, "Unable to parse '%s'", orig);
					// reset back to the default value
					keywords[i].handler(keywords[i].def, keywords[i].var);
				}
	}
	fclose(in);
	return (1);
}

#ifdef CYGPKG_ELIOSDHCP_LEASE_FILE
//-----------------------------------------------------------------------------------------
// Write lease File
// @Return 0 on success , -1 otherwise
//-----------------------------------------------------------------------------------------
char write_leases(void)
{
	FILE *fp=NULL;
	unsigned int i=0;
	time_t curr = time(0);
	unsigned long lease_time=0;

	DEBUG("write_leases","Called");
	if (!(fp = fopen(server_config.lease_file, "w"))) {
		LOG(LOG_ERR,"write_leases","Unable to open leases file '%s' for writing : '%s'", server_config.lease_file,strerror(errno));
		return(-1);
	}

	for (i = 0; i < server_config.max_leases; i++)
	{
		if (leases[i].yiaddr != 0)
		{
			if (server_config.remaining)
			{
				if (lease_expired(&(leases[i])))
					lease_time = 0;
				else lease_time = leases[i].expires - curr;
			} else lease_time = leases[i].expires;
			lease_time = htonl(lease_time);
			fwrite(leases[i].chaddr,TABLELEASESIZE, 1, fp);
			fwrite(&(leases[i].yiaddr), 4, 1, fp);
			fwrite(&lease_time, sizeof(lease_time), 1, fp);
		}
	}
	fclose(fp);
	return(0);
}

//-----------------------------------------------------------------------------------------
// Read lease File
// @Return : -1 on error, number of read_leases otherwise
//-----------------------------------------------------------------------------------------
int read_leases(char *file)
{
	FILE *fp=NULL;
	unsigned int i = 0;
	struct dhcpOfferedAddr lease;

	DEBUG("read_leases","Called");
	if (!(fp=fopen(file, "r")))
	{
		LOG(LOG_ERR,"read_leases","Unable to open leases file '%s' for reading : %s", file,strerror(errno));
		return(-1);
	}
	while (i < server_config.max_leases && (fread(&lease, sizeof (lease), 1, fp) == 1))
	{
		if (lease.yiaddr >= server_config.start && lease.yiaddr <= server_config.end)
		 {
			lease.expires = ntohl(lease.expires);
			if (!server_config.remaining) lease.expires -= time(0);
			if (!(add_lease(lease.chaddr, lease.yiaddr, lease.expires,TABLELEASESIZE)))
			{
				LOG(LOG_WARNING,"read_leases","Too many leases while loading %s", file);
				break;
			}
			i++;
		}
	}
	LOG(LOG_INFO,"read_leases","Read %d leases", i);
	fclose(fp);
	return(i);
}
#endif

