/*
 * leases.c -- tools to manage DHCP leases
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <time.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "ecos.h"
#include "files.h"
#include "options.h"
#include <net/dhcpelios/debug.h>
#include <net/dhcpelios/leases.h>
#include <net/dhcpelios/dhcpd.h>

unsigned char blank_chaddr[] = {[0 ... TABLELEASESIZE-1] = 0};

//-------------------------------------------------------------------------------
// Clear every lease out that chaddr OR yiaddr matches and is nonzero
// @param
//   chaddr is an array containing MAC address
//   yiaddr is the IP address
//   siz is size of a chaddr
// @return : nothing
//-------------------------------------------------------------------------------
void clear_lease(u_int8_t *chaddr, u_int32_t yiaddr,unsigned int siz)
{
	unsigned int i=0;		// Lease index
	unsigned int j=0;

	DEBUG("clear_lease","Called");
	// Jump over zeroed chaddr
	for (j = 0; j < siz && !chaddr[j]; j++);

	for (i = 0; i < server_config.max_leases; i++)
	 if ((j != siz && !memcmp(leases[i].chaddr, chaddr, siz)) || (yiaddr && leases[i].yiaddr == yiaddr))
	 {
	  memset(&(leases[i]), 0, sizeof(struct dhcpOfferedAddr)); // clear lease
	 }
}

//-------------------------------------------------------------------------------
// Add a lease into the table, clearing out any old ones
// could be use to modify packet !
// @param
//   chaddr is an array containing MAC address
//   yiaddr is the IP address
//   lease is the lease time
//   siz is size of a chaddr
// @return
//  pointer on the new lease
//-------------------------------------------------------------------------------
struct dhcpOfferedAddr *add_lease(u_int8_t *chaddr, u_int32_t yiaddr, unsigned long lease,unsigned int siz)
{
	struct dhcpOfferedAddr *oldest=NULL;
#ifdef DEBUGGING
 	char ip[IPSTRING];
#endif

	DEBUG("add_lease","Called with %s siz:%d and lease time:%ul",DHCPNumber2String(yiaddr,ip),siz,lease);
	// clean out any old ones
	clear_lease(chaddr, yiaddr,siz);
	// Scan oldest expired one
	oldest = oldest_expired_lease();
	if (oldest)
	{
		memcpy(oldest->chaddr, chaddr,siz);	// Copy MAC address
		oldest->yiaddr = yiaddr;		// IP adress
		oldest->expires = time(0) + lease;	// New Leased time
	}
	return (oldest);
}

//-------------------------------------------------------------------------------
// @param lease : ponter on a lease
// @return true if a lease has expired, false otherwise
//-------------------------------------------------------------------------------
char lease_expired(struct dhcpOfferedAddr *lease)
{
 DEBUG("lease_expired","Called");
 return (lease->expires < (unsigned long) time(0));
}


//-------------------------------------------------------------------------------
// @return the oldest expired lease adress, NULL if there are no expired leases
//-------------------------------------------------------------------------------
struct dhcpOfferedAddr *oldest_expired_lease(void)
{
	struct dhcpOfferedAddr *oldest = NULL;
	unsigned long oldest_lease = time(0);
	unsigned int i=0;

	DEBUG("oldest_expired_lease","Called");
	for (i = 0; i < server_config.max_leases; i++)
		if (oldest_lease > leases[i].expires) {
			oldest_lease = leases[i].expires;
			oldest = &(leases[i]);
		}
	return (oldest);
}

//-------------------------------------------------------------------------------
// Find the first lease that matches MAC address chaddr, NULL if no match
// @param chaddr : MAC address
// @return
//  pointer on the lease or NULL
//-------------------------------------------------------------------------------
struct dhcpOfferedAddr *find_lease_by_chaddr(u_int8_t *chaddr)
{
#ifdef DEBUGGING
 char arp[MACSTRING]
#endif
	unsigned int i=0;
	DEBUG("Find_lease_by_chaddr","Called with '%s'",DHCPether_print(chaddr,arp,MACSTRING));
	for (i = 0; i < server_config.max_leases; i++)
		if (!memcmp(leases[i].chaddr, chaddr, TABLELEASESIZE)) return &(leases[i]);
	return (NULL);
}


//-------------------------------------------------------------------------------
// Find the first lease that matches yiaddr(IP), NULL is no match
// @param yiaddr : IP address
// @return
//  pointer on the lease or NULL
//-------------------------------------------------------------------------------
struct dhcpOfferedAddr *find_lease_by_yiaddr(u_int32_t yiaddr)
{
	unsigned int i=0;
#ifdef DEBUGGING
 	char ip[IPSTRING];
#endif

	DEBUG("Find_lease_by_yiaddr","Called with %s",DHCPNumber2String(yiaddr,ip));
	for (i = 0; i < server_config.max_leases; i++)
		if (leases[i].yiaddr == yiaddr) return &(leases[i]);
	return (NULL);
}


//-------------------------------------------------------------------------------
// Find an assignable address, it check_expired is true, we check all the expired leases as well.
// Maybe this should try expired leases by age...
// @param  check_expired : boolean to seek also expirated ones
// @return IP address
//            0 is no match
//-------------------------------------------------------------------------------
u_int32_t find_address(char check_expired)
{
	u_int32_t addr=0;
	u_int32_t ret=0;
	struct dhcpOfferedAddr *lease = NULL;

	DEBUG("Find address","Called with check_expired=%d",check_expired);
	addr = ntohl(server_config.start); // addr is in host order here
	for (;addr <= ntohl(server_config.end); addr++)
	{
		// ie, 192.168.55.0
		if (!(addr & 0xFF)) continue; // Network address

		// ie, 192.168.55.255
		if ((addr & 0xFF) == 0xFF) continue; // Broadcast address

		// lease is not taken
		ret = htonl(addr);			// Network order
		if ((!(lease = find_lease_by_yiaddr(ret)) ||

		     // or it expired and we are checking for expired leases
		     (check_expired  && lease_expired(lease))) &&

		     // and it isn't on the network
	    	     !check_ip(ret))
		{
		 return (ret);
		}
	}
	return (0);
}

//-------------------------------------------------------------------------------
// Check is an IP is taken, if it is, add it to the lease table
// @param addr : IP address
// @Return 1 if address is used 0 otherwise
//-------------------------------------------------------------------------------
char check_ip(u_int32_t addr)
{
        char ip[IPSTRING];					// IP address conversion

	DHCPNumber2String(addr,ip);
	DEBUG("check_ip","Called with '%s'",ip);
	if (DHCPPingHost(ip,1)==1)
	{
	 	LOG(LOG_ERR,"check_ip","%s belongs to someone, reserving it for %ld seconds",ip, server_config.conflict_time);
		add_lease(blank_chaddr, addr, server_config.conflict_time,TABLELEASESIZE);
		return (1);
	}
	else
	 LOG(LOG_INFO,"check_ip","%s belongs to no one ((doesn't answer to ping)",ip);
	return (0);
}

