/*
 * options.c -- DHCP server client/server options
 *
 * udhcp client/server
 * options.c -- DHCP server option packet tools
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <network.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>

#include "files.h"
#include "options.h"
#include <net/dhcpelios/leases.h>
#include <net/dhcpelios/debug.h>
#include <net/dhcpelios/dhcpd.h>


// Supported options are easily added here
struct dhcp_option options[] = {
	{"subnet",	OPTION_IP | OPTION_REQ,			0x01},
	{"timezone",	OPTION_S32,				0x02},
	{"router",	OPTION_IP | OPTION_REQ,			0x03},	// OPTION_LIST removed
	{"timesvr",	OPTION_IP | OPTION_LIST,		0x04},
	{"namesvr",	OPTION_IP | OPTION_LIST,		0x05},
	{"dns",		OPTION_IP | OPTION_REQ,			0x06},  // OPTION_LIST removed
 	{"logsvr",	OPTION_IP | OPTION_LIST,		0x07},
	{"cookiesvr",	OPTION_IP | OPTION_LIST,		0x08},
	{"lprsvr",	OPTION_IP | OPTION_LIST,		0x09},
	{"hostname",	OPTION_STRING | OPTION_REQ,		0x0c},
	{"bootsize",	OPTION_U16,				0x0d},
	{"domain",	OPTION_STRING | OPTION_REQ,		0x0f},
	{"swapsvr",	OPTION_IP,				0x10},
	{"rootpath",	OPTION_STRING,				0x11},
	{"ipttl",	OPTION_U8,				0x17},
	{"mtu",		OPTION_U16,				0x1a},
	{"broadcast",	OPTION_IP | OPTION_REQ,			0x1c},
	{"ntpsrv",	OPTION_IP | OPTION_LIST,		0x2a},
	{"wins",	OPTION_IP | OPTION_LIST,		0x2c},
	{"requestip",	OPTION_IP,				0x32},
	{"lease",	OPTION_U32,				0x33},
	{"dhcptype",	OPTION_U8,				0x35},
	{"serverid",	OPTION_IP,				0x36},
	{"message",	OPTION_STRING,				0x38},
	{"tftp",	OPTION_STRING,				0x42},
	{"bootfile",	OPTION_STRING,				0x43},
	{"",		0x00,				0x00}
};

// Length in bytes of the different option types
int option_lengths[] = {
	[OPTION_IP] =		4,
	[OPTION_IP_PAIR] =	8,
	[OPTION_BOOLEAN] =	1,
	[OPTION_STRING] =	1,
	[OPTION_U8] =		1,
	[OPTION_U16] =		2,
	[OPTION_S16] =		2,
	[OPTION_U32] =		4,
	[OPTION_S32] =		4
};

//---------------------------------------------------------------------------------------
// Get an option in packet with bounds checking (warning, not aligned).
// code should be : DHCP_REQUESTED_IP,DHCP_LEASE_TIME,DHCP_VENDOR,DHCP_MESSAGE_TYPE,DHCP_SERVER_ID
// Returns NULL on error or a pointer on the value
//---------------------------------------------------------------------------------------
unsigned char *get_option(struct dhcpMessage *packet, int code)
{
	int i=0;				// Index
	int length=OPTSIZE;			// Maximum size of an option
	unsigned char *optionptr=NULL;		// Pointer on current option
	char over = 0;				// Boolean
	char done = 0;				// Boolean
	int curr = OPTION_FIELD;

	DEBUG("get_option","Called");
	optionptr = packet->options; // This is an array of OPTSIZE bytes
	while (!done)
	{
		if (i >= length)
		{
		 LOG(LOG_WARNING,"get_option","Bogus packet, we have reach end of options");
		 return (NULL);
		}
		if (optionptr[i + OPT_CODE] == code) // Is this the option we were searching ?
		{
		 if (i + 1 + optionptr[i + OPT_LEN] >= length) // End of data after maximum length ?
		 {
		  LOG(LOG_WARNING,"get_option","Bogus packet, option field too long.");
		  return (NULL);
		 }
		 return (optionptr + i + OPT_DATA); // Return pointer on value
		}

		switch (optionptr[i + OPT_CODE])
		{
		 case DHCP_PADDING:
			i++; // go to next
			break;
		 case DHCP_OPTION_OVER:
			if (i + 1 + optionptr[i + OPT_LEN] >= length) // End of data after maximum length ?
			{
				LOG(LOG_WARNING,"get_option","Bogus  packet, option OVER field too long.");
				return (NULL);
			}
			over = optionptr[i + 3];
			i += optionptr[i+OPT_LEN] + 2; // Next option
			break;
		 case DHCP_END:
			if (curr == OPTION_FIELD && over & FILE_FIELD) {
				optionptr = packet->file;
				i = 0;
				length = sizeof(packet->file);
				curr = FILE_FIELD;
			} else if (curr == FILE_FIELD && over & SNAME_FIELD) {
				optionptr = packet->sname;
				i = 0;
				length = sizeof(packet->sname);
				curr = SNAME_FIELD;
			} else done = 1;
			break;
		 default:
			i += optionptr[OPT_LEN + i] + 2; // Next option
		}
	}
	return (NULL);
}

//---------------------------------------------------------------------------------------
// Return the position of the 'end' option with bounds checking
//---------------------------------------------------------------------------------------
int end_option(unsigned char *optionptr)
{
	int i = 0;
	int length=OPTSIZE;			// Maximum size of an option

	DEBUG("end_option","Called");
	while ((optionptr[i] != DHCP_END)&&(i<length)) {
		if (optionptr[i] == DHCP_PADDING) i++;
		else i += optionptr[i + OPT_LEN] + 2; // Next option
	}
	if (i<length)
	 return (i);


	else return(-1);
}


//---------------------------------------------------------------------------------------
// Add an option string to the options beggining at optionptr
// (An option string contains an option code,length, then data)
// Returns 0 on error, sinon la nouvelle position de la fin
//---------------------------------------------------------------------------------------
int add_option_string(unsigned char *optionptr, unsigned char *string)
{
	int end = end_option(optionptr); // Search for end

	DEBUG("add_option_string","Called");
	if (end<0) return(-1);		 // No end found ?
	// end position + string length + option code/length + end option
	if (end + string[OPT_LEN] + 2 + 1 >= OPTSIZE) {
		LOG(LOG_ERR,"add_option_string","Option 0x%02x did not fit into the packet!", string[OPT_CODE]);
		return (0);
	}
	LOG(LOG_INFO,"add_option_string","Adding option 0x%02x", string[OPT_CODE]);
	memcpy(optionptr + end, string, string[OPT_LEN] + 2);
	optionptr[end + string[OPT_LEN] + 2] = DHCP_END;
	return (string[OPT_LEN] + 2);
}

//---------------------------------------------------------------------------------------
// Add a one to four byte option to a packet to the options beggining at optionptr
// code is the id of option, data is the value
// Returns 0 on error
//---------------------------------------------------------------------------------------
int add_simple_option(unsigned char *optionptr, unsigned char code, u_int32_t data)
{
	char length=0;
	int i=0;			// Option scanning index
	unsigned char option[2 + 4];
	unsigned char *u8=NULL;
	u_int16_t *u16=NULL;
	u_int32_t *u32=NULL;
	u_int32_t aligned=0;
	u8 = (unsigned char *) &aligned;
	u16 = (u_int16_t *) &aligned;
	u32 = &aligned;

	DEBUG("add_simple_option","Called");
	for (i = 0; options[i].code; i++) // Scanning known options
	 if (options[i].code == code)
	  length = option_lengths[options[i].flags & TYPE_MASK]; // Get size of this option

	if (!length) {
		LOG(LOG_ERR,"add_simple_option","Could not add option 0x%02x", code);
		return (0);
	}

	option[OPT_CODE] = code;
	option[OPT_LEN] = length;

	switch (length) {
		case 1: *u8 =  data; break;
		case 2: *u16 = data; break;
		case 4: *u32 = data; break;
	}
	memcpy(option + OPT_DATA, &aligned, length);
	return (add_option_string(optionptr, option));
}

//---------------------------------------------------------------------------------------
// find option 'code' in opt_list (chained list)
// Return NULL if not found or a pointer on option
// code can be :
// DHCP option codes (partial list)
// DHCP_PADDING		0x00
// DHCP_SUBNET		0x01
// DHCP_TIME_OFFSET	0x02
// DHCP_ROUTER		0x03
// DHCP_TIME_SERVER	0x04
// DHCP_NAME_SERVER	0x05
// DHCP_DNS_SERVER	0x06
// DHCP_LOG_SERVER	0x07
// DHCP_COOKIE_SERVER	0x08
// DHCP_LPR_SERVER	0x09
// DHCP_HOST_NAME	0x0c
// DHCP_BOOT_SIZE	0x0d
// DHCP_DOMAIN_NAME	0x0f
// DHCP_SWAP_SERVER	0x10
// DHCP_ROOT_PATH	0x11
// DHCP_IP_TTL		0x17
// DHCP_MTU		0x1a
// DHCP_BROADCAST	0x1c
// DHCP_NTP_SERVER	0x2a
// DHCP_WINS_SERVER	0x2c
// DHCP_REQUESTED_IP	0x32
// DHCP_LEASE_TIME	0x33
// DHCP_OPTION_OVER	0x34
// DHCP_MESSAGE_TYPE	0x35
// DHCP_SERVER_ID	0x36
// DHCP_PARAM_REQ	0x37
// DHCP_MESSAGE		0x38
// DHCP_MAX_SIZE	0x39
// DHCP_T1		0x3a
// DHCP_T2		0x3b
// DHCP_VENDOR		0x3c
// DHCP_CLIENT_ID	0x3d
// DHCP_END		0xFF
//---------------------------------------------------------------------------------------
struct option_set *find_option(struct option_set *opt_list, char code)
{
	DEBUG("find_option","Called");
	while (opt_list && opt_list->data[OPT_CODE] < code)
		opt_list = opt_list->next; // scan the sorted list

	if (opt_list && opt_list->data[OPT_CODE] == code) return (opt_list);
	else return (NULL);
}

//---------------------------------------------------------------------------------------
// Add an option to the opt_list
//---------------------------------------------------------------------------------------
void attach_option(struct option_set **opt_list, struct dhcp_option *option, char *buffer, int length)
{
	struct option_set *existing=NULL;
	struct option_set *new=NULL;
	struct option_set **curr=NULL;

	DEBUG("attach_option","Called");
	// Add it to an existing option
	if ((existing = find_option(*opt_list, option->code))) {
		LOG(LOG_INFO,"attach_option","Attaching option %s to existing member of list", option->name);
		if (option->flags & OPTION_LIST) {
			if (existing->data[OPT_LEN] + length <= 255) {
				existing->data = realloc(existing->data,
						existing->data[OPT_LEN] + length + 2);
				memcpy(existing->data + existing->data[OPT_LEN] + 2, buffer, length);
				existing->data[OPT_LEN] += length;
			} // else, ignore the data, we could put this in a second option in the future
		} // else, ignore the new data
	} else {
		LOG(LOG_INFO,"attach_option","Attaching option %s to list", option->name);

		// make a new option
		new = malloc(sizeof(struct option_set));
		new->data = malloc(length + 2);
		new->data[OPT_CODE] = option->code;
		new->data[OPT_LEN] = length;
		memcpy(new->data + 2, buffer, length);

		curr = opt_list;
		while (*curr && (*curr)->data[OPT_CODE] < option->code)
			curr = &(*curr)->next;

		new->next = *curr;
		*curr = new;
	}
}

//---------------------------------------------------------------------------------------
// Free the memory used by option list
//---------------------------------------------------------------------------------------
void FreeOptionsList(struct option_set *opt_list)
{
 struct option_set *curr=NULL;
 struct option_set *suiv=NULL;

 DEBUG("FreeOptionsList","Called with %p", opt_list);
 curr = opt_list;
 while (curr)
 {
  suiv = curr->next;
  if (curr->data)
  {
    LOG(LOG_INFO,"FreeOptionsList","Freeing %p option :%02x", curr,curr->data[OPT_CODE]);
   free(curr->data);
  }
  free(curr);
  curr=suiv;
 }
 opt_list=NULL;
}
