/*
 * packet.c -- DHCP server client/server packet creation
 *
 * udhcp client/server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <network.h>
#include <net/ethernet.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_var.h>
#include "ecos.h"
#include "packet.h"
#include "options.h"
#include <net/dhcpelios/debug.h>
#include <net/dhcpelios/dhcpd.h>

//------------------------------------------------------------------------------------------------------
// Initialize DHCP header (in buffer packet) .
// type is is DHCP type of the packet (DHCPDISCOVER,DHCPREQUEST,DHCPRELEASE,DHCPINFORM,DHCPOFFER,DHCPACK,DHCPNAK)
// Returns nothing
//------------------------------------------------------------------------------------------------------
void init_header(struct dhcpMessage *packet, char type)
{
	DEBUG("init_header","Called with type %02x",type);
	memset(packet, 0, sizeof(struct dhcpMessage));
	switch (type)
	{
	 case DHCPDISCOVER:
	 case DHCPREQUEST:
	 case DHCPRELEASE:
	 case DHCPINFORM:
		packet->op = BOOTREQUEST; // request
		break;
	 case DHCPOFFER:
	 case DHCPACK:
	 case DHCPNAK:
		packet->op = BOOTREPLY; // Response
	}
	packet->htype = ETH_10MB;	// Hardware type
	packet->hlen = ETH_10MB_LEN;	// Hardware address len
	packet->cookie = htonl(DHCP_MAGIC); // ID MAGIC
	packet->options[0] = DHCP_END;	// Fin des options
	add_simple_option(packet->options, DHCP_MESSAGE_TYPE, type);
}

//------------------------------------------------------------------------------------------------------
// read a DHCP packet from socket fd.
// @return -1 on read error, -2 on packet error or number of read bytes
//------------------------------------------------------------------------------------------------------
int get_packet(struct dhcpMessage *packet, int fd)
{
	int bytes=0;
	int i=0;
	const char broken_vendors[][8] = {
		"MSFT 98",
		""
	};
	char unsigned *vendor=NULL;
	char tmpvendor[64];
#ifdef LOGACTIVE
 char ip[IPSTRING];
#endif

	DEBUG("get_packet","Called with socket %d",fd);
	memset(packet, 0, sizeof(struct dhcpMessage));
	bytes = read(fd, packet, sizeof(struct dhcpMessage));
	if (bytes < 0)
	{
	 LOG(LOG_ERR,"get_packet","Couldn't read on listening socket, ignoring :'%s'",strerror(errno));
	 return (-1);
	}

	LOG(LOG_INFO,"get_packet", "Received a packet of %d/%d bytes",bytes,sizeof(struct dhcpMessage));
	LOG(LOG_INFO,"get_packet", "Op:%d Htype :%d",packet->op,packet->htype);
	LOG(LOG_INFO,"get_packet", "Hlen:%d Hops:%d",packet->hlen,packet->hops);
	LOG(LOG_INFO,"get_packet", "Xid:%08x Secs:%d",packet->xid,packet->secs);
	LOG(LOG_INFO,"get_packet", "Flags:%d Cookie:%08x",packet->flags,packet->cookie);
	LOG(LOG_INFO,"get_packet", "ciaddr:%s",DHCPNumber2String(packet->ciaddr,ip));
	LOG(LOG_INFO,"get_packet", "yiaddr:%s",DHCPNumber2String(packet->yiaddr,ip));
	LOG(LOG_INFO,"get_packet", "siaddr:%s",DHCPNumber2String(packet->siaddr,ip));
	LOG(LOG_INFO,"get_packet", "giaddr:%s",DHCPNumber2String(packet->giaddr,ip));

	if (ntohl(packet->cookie) != DHCP_MAGIC)
	{
	 LOG(LOG_ERR,"get_packet","Invalid DHCP cookie");
	 return (-2);
	}

	if (packet->op == BOOTREQUEST && (vendor = get_option(packet, DHCP_VENDOR)))
	{
		strncpy(tmpvendor,vendor,sizeof(tmpvendor)<vendor[OPT_LEN - 2]?sizeof(tmpvendor):vendor[OPT_LEN - 2]);
		for (i = 0; broken_vendors[i][0]; i++)
		{
			LOG(LOG_INFO,"get_packet", "Scanning vendors :'%s'/'%s' %d",broken_vendors[i],tmpvendor,vendor[OPT_LEN - 2]);
			if (vendor[OPT_LEN - 2] == (unsigned char) strlen(broken_vendors[i]) &&
			    !strncmp(vendor, broken_vendors[i], vendor[OPT_LEN - 2])) {
			    	LOG(LOG_WARNING,"get_packet", "Broken client ('%s'), forcing broadcast",broken_vendors[i]);
			    	packet->flags |= htons(BROADCAST_FLAG);
			}
		}
	}
	return (bytes);
}

/*
//------------------------------------------------------------------------------------------------------
// Compute Internet Checksum for "count" bytes beginning at location "addr".
//------------------------------------------------------------------------------------------------------
u_int16_t checksum(void *addr, int count)
{
	register int32_t sum = 0;
	u_int16_t *source = (u_int16_t *) addr;

	DEBUG("checksum","Called");
	while (count > 1)  {
		//  This is the inner loop
		sum += *source++;
		count -= 2;
	}

	//  Add left-over byte, if any
	if (count > 0) {
		// Make sure that the left-over byte is added correctly both
		// with little and big endian hosts
		u_int16_t tmp = 0;
		*(unsigned char *) (&tmp) = * (unsigned char *) source;
		sum += tmp;
	}
	//  Fold 32-bit sum to 16 bits
	while (sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);

	return (~sum);
}
*/

//------------------------------------------------------------------------------------------------------
// Construct an ip/udp header for a packet, and specify the source(source_ip,source_port) and destination(dest_ip,dest_port,dest_arp,ifindex)
// @Returns number of sent bytes
//------------------------------------------------------------------------------------------------------
int raw_packet(struct dhcpMessage *payload, u_int32_t source_ip, int source_port,
		   u_int32_t dest_ip, int dest_port, unsigned char *dest_arp, int ifindex)
{
	struct sockaddr_in dest_host;		// Target
	int ret=0;				// Returned value
	int retmac=0;				// Return value for DHCPAddMAC

#ifdef DEBUGGING
 char ip[IPSTRING];
 char ip2[IPSTRING];
 char arp[MACSTRING];
 DHCPether_print(dest_arp,arp,MACSTRING);
#endif


	DEBUG("raw_packet","Called from :'%s':%d  to :'%s':%d [%s] by interface %d",DHCPNumber2String(source_ip,ip),source_port,DHCPNumber2String(dest_ip,ip2),dest_port,arp,ifindex);

	memset(&dest_host, 0, sizeof(dest_host));
	dest_host.sin_family = AF_INET;
	dest_host.sin_port = htons(dest_port);
	dest_host.sin_addr.s_addr = dest_ip;
	dest_host.sin_len=sizeof(dest_host);

	retmac=DHCPAddMAC(inet_ntoa(dest_host.sin_addr),dest_arp,ifindex);		// Add temp mac entry
	ret=kernel_packet(payload,source_ip,source_port,dest_ip,dest_port);		// Send packet
	if (retmac==0) DHCPDelMAC(inet_ntoa(dest_host.sin_addr),dest_arp,ifindex);      // Suppress temp MAC entry
	return(ret);
}

//------------------------------------------------------------------------------------------------------
// Let the kernel do all the work for packet generation
// Send a packet (payload) from source (source_ip,source_port) to destination (dest_ip,dest_port)
// @Returns -1 on error or number of sent bytes
//------------------------------------------------------------------------------------------------------
int kernel_packet(struct dhcpMessage *payload, u_int32_t source_ip, int source_port,
		   u_int32_t dest_ip, int dest_port)
{
	int n=1;			 // Activate REUSEADDR
	int fd=0;			 // Socket file descriptor
	int result=0;			 // Result of send
	struct sockaddr_in client;	 // Socket adress struct for target
	struct sockaddr_in source;       // Socket adress struct for source
	struct udp_dhcp_packet packet;   // for sending in broadcast mode
#ifdef DEBUGGING
 char ip[IPSTRING];
 char ip2[IPSTRING];
#endif

	DEBUG("kernel_packet","Called from :'%s':%d  to :'%s':%d",DHCPNumber2String(source_ip,ip),source_port,DHCPNumber2String(dest_ip,ip2),dest_port);

	memset(&source, 0, sizeof(source));
	source.sin_family = AF_INET;
	source.sin_addr.s_addr = source_ip;
	source.sin_len=sizeof(source);
	source.sin_port = htons(source_port);

	memset(&client, 0, sizeof(client));
	client.sin_family = AF_INET;
	client.sin_addr.s_addr = dest_ip;
	client.sin_len=sizeof(client);
        client.sin_port = htons(dest_port);

	if (dest_ip!=INADDR_BROADCAST)
	{
	 // Unicast mode
         LOG(LOG_INFO,"kernel_packet", "Broadcasting not required");
  	 if ((fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	 {
 	  LOG(LOG_ERR,"kernel_packet","Cannot obtain UDP socket:'%s'",strerror(errno));
 	  return (-1);
	 }
	 if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &n, sizeof(n)) == -1)
	 {
	  close(fd);
 	  LOG(LOG_ERR,"kernel_packet","Cannot REUSEADDR via %d : '%s'",fd,strerror(errno));
	  return (-1);
	 }
	 if (bind(fd, (struct sockaddr *)&source, sizeof(struct sockaddr)) == -1)
	 {
	  close(fd);
 	  LOG(LOG_ERR,"kernel_packet","Cannot bind to '%s':%d via %d : '%s'",inet_ntoa(source.sin_addr),ntohs(source.sin_port),fd,strerror(errno));
	  return (-1);
	 }
	 else LOG(LOG_INFO,"kernel_packet","Bind to '%s':%d via %d",inet_ntoa(source.sin_addr),ntohs(source.sin_port),fd);
	 if (connect(fd, (struct sockaddr *)&client, sizeof(struct sockaddr)) == -1)
	 {
	  close(fd);
	  LOG(LOG_ERR,"kernel_packet","Cannot connect to '%s':%d via socket no %d : '%s'",inet_ntoa(client.sin_addr),ntohs(client.sin_port),fd,strerror(errno));
	  return (-1);
	 }
	 // Send packet
	 result = write(fd, payload, sizeof(struct dhcpMessage));
	}
	else
	{
	 // Broadcast mode
	 LOG(LOG_INFO,"kernel_packet", "Broadcasting required");
	 if ((fd = socket(AF_INET, SOCK_RAW, IPPROTO_UDP)) < 0)
	 {
 	  LOG(LOG_ERR,"kernel_packet","Cannot obtain raw UDP socket:'%s'",strerror(errno));
 	  return (-1);
	 }
         packet.udp.uh_sport = htons(source_port);
	 packet.udp.uh_dport = htons(dest_port);
         packet.udp.uh_ulen = htons(sizeof(packet.udp) + sizeof(struct dhcpMessage)); // cheat on the pseudo-header
         memcpy(&(packet.data), payload, sizeof(struct dhcpMessage));
         packet.udp.uh_sum =0; // errnoneous UDP checksum
	 //checksum(&packet, sizeof(struct udp_dhcp_packet));

         result=sendto(fd,&packet, sizeof(struct udp_dhcp_packet),0,(struct sockaddr *)&client,sizeof(client));
	}
	if (result <= 0)
	{
	 LOG(LOG_ERR,"kernel_packet","Write on socket no %d failed: '%s'",fd,strerror(errno));
	}
	else LOG(LOG_INFO,"kernel_packet","Wrote %d bytes to '%s'[%x]:%d via socket no %d",result,inet_ntoa(client.sin_addr),client.sin_addr.s_addr,ntohs(client.sin_port),fd);
	close(fd);
	return (result);
}


