/* serverpacket.c
 *
 * Constuct and send DHCP server packets
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <time.h>

#include "packet.h"
#include "options.h"
#include "ecos.h"
#include <net/dhcpelios/leases.h>
#include <net/dhcpelios/dhcpd.h>
#include <net/dhcpelios/debug.h>

//------------------------------------------------------------------------------------------------------
// Send a packet to giaddr using the kernel ip stack
//------------------------------------------------------------------------------------------------------
static int send_packet_to_relay(struct dhcpMessage *payload)
{
#ifdef LOGACTIVE
 char gw[IPSTRING];	// IP adress of gateway in human readable form
#endif

 DEBUG("send_packet_to_relay","Called");
 LOG(LOG_INFO,"send_packet_to_relay","Forwarding packet to relay :%s",DHCPNumber2String(payload->giaddr,gw));
 return kernel_packet(payload, server_config.server,SERVER_PORT,payload->giaddr,SERVER_PORT);
}

//------------------------------------------------------------------------------------------------------
// send a packet to a specific arp address and ip address by creating our own ip packet
// @param force_broadcast is a boolean to activate broadcast
//------------------------------------------------------------------------------------------------------
static int send_packet_to_client(struct dhcpMessage *payload,char force_broadcast)
{
	unsigned char *chaddr=NULL;	// MAC adress of target
	u_int32_t ciaddr=0;		// IP adress of target
#ifdef LOGACTIVE
	char ip[IPSTRING];		// IP adress conversion
	char arp[MACSTRING];		// MAC adress conversion
#endif

	DEBUG("send_packet_to_client","Called with force_broadcast:%d",force_broadcast);
	if ((force_broadcast)||(ntohs(payload->flags) & BROADCAST_FLAG))
	{
		LOG(LOG_INFO,"send_packet_to_client","Broadcasting packet to client)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else if (payload->ciaddr)
	{
		chaddr = payload->chaddr;
		ciaddr = payload->ciaddr;
		LOG(LOG_INFO,"send_packet_to_client","Unicasting packet to client ciaddr '%s' [%s]",DHCPNumber2String(ciaddr,ip),DHCPether_print(chaddr,arp,MACSTRING));
	} else
	{
		ciaddr = payload->yiaddr;
		chaddr = payload->chaddr;
		LOG(LOG_INFO,"send_packet_to_client","Unicasting packet to client yiaddr '%s' [%s]",DHCPNumber2String(ciaddr,ip),DHCPether_print(chaddr,arp,MACSTRING));
	}
	return (raw_packet(payload, server_config.server, SERVER_PORT,ciaddr, CLIENT_PORT, chaddr, server_config.ifindex));
}

//------------------------------------------------------------------------------------------------------
// Send a dhcp packet, if force broadcast is set, the packet will be broadcasted to the client
// @param force_broadcast is a boolean to activate broadcast
// @Returns number of sent bytes or -1
//------------------------------------------------------------------------------------------------------
static int send_packet(struct dhcpMessage *payload,char force_broadcast)
{
	int ret=0;	// number of sent bytes or -1

	DEBUG("send_packet","Called with force_broadcast :%d",force_broadcast);

	if (payload->giaddr)
		ret = send_packet_to_relay(payload);
	else ret = send_packet_to_client(payload, force_broadcast);
	return (ret);
}

//------------------------------------------------------------------------------------------------------
// Init a DHCP packet with an old one
//------------------------------------------------------------------------------------------------------
static void init_packet(struct dhcpMessage *packet, struct dhcpMessage *oldpacket, char type)
{
	DEBUG("init_packet","Called with type :%02x",type);

	init_header(packet, type);
	packet->xid = oldpacket->xid;
	LOG(LOG_INFO,"init_packet","ID request :%08x",packet->xid);
	memcpy(packet->chaddr, oldpacket->chaddr,sizeof(packet->chaddr)); // Copy target MAC adress
	packet->flags = oldpacket->flags;
	packet->giaddr = oldpacket->giaddr;	       // Copy gateway adress
	packet->ciaddr = oldpacket->ciaddr;
	add_simple_option(packet->options, DHCP_SERVER_ID, server_config.server); // set server IP
}

//------------------------------------------------------------------------------------------------------
// Add in the bootp options
//------------------------------------------------------------------------------------------------------
static void add_bootp_options(struct dhcpMessage *packet)
{
	DEBUG("add_bootp_options","Called");

	packet->siaddr = server_config.siaddr;   // IP Address of DHCP server
	if (server_config.sname)
		strncpy(packet->sname, server_config.sname, sizeof(packet->sname) - 1); // Server name
	if (server_config.boot_file)
		strncpy(packet->file, server_config.boot_file, sizeof(packet->file) - 1); // Boot file
}

//------------------------------------------------------------------------------------------------------
// Add in the server options (except LEASE_TIME)
//------------------------------------------------------------------------------------------------------
static void add_server_options(struct dhcpMessage *packet)
{
 struct option_set *curr=NULL;			// Current server option

 DEBUG("add_server_options","Called");
 // Add all server,options except LEASE_TIME
 curr = server_config.options;
 while (curr)
 {
  if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
   add_option_string(packet->options, curr->data);
  curr = curr->next;
 }
}

//------------------------------------------------------------------------------------------------------
// Send a DHCP OFFER to a answer DHCP DISCOVER
// Returns number of bytes sent or -1
//------------------------------------------------------------------------------------------------------
int sendOffer(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;			// New packet
	struct dhcpOfferedAddr *lease = NULL;		// Choosed lease
	u_int32_t req_align=0;				// REQUESTED_IP
	u_int32_t lease_time_align = server_config.lease;// LEASE_TIME
	unsigned char *req=NULL;			// Result of get_option
	unsigned char *lease_time=NULL;			// LEASE_TIME
#ifdef LOGACTIVE
	char ip[IPSTRING];				// IP adress conversion
#endif


	DEBUG("sendOffer","Called");

	init_packet(&packet, oldpacket, DHCPOFFER); // Compute a DHCPOFFER packet

	// The MAC adress of client is in our lease table ?
	if ((lease = find_lease_by_chaddr(oldpacket->chaddr)))
        {
		if (!lease_expired(lease)) // Has it expired ?
			lease_time_align = lease->expires - time(0);  // Remaining time
		packet.yiaddr = lease->yiaddr; // Giving known IP
		LOG(LOG_INFO,"sendOffer","Choosed to offer adress '%s' which is in our lease table",DHCPNumber2String(packet.yiaddr,ip));
	}
	else
	{
    	  // The client has a requested ip
          req = get_option(oldpacket, DHCP_REQUESTED_IP);
	  if (req)
	  {
   	   memcpy(&req_align, req,sizeof(req_align));
	   LOG(LOG_INFO,"sendOffer","Client wants :'%s'",DHCPNumber2String(req_align,ip));
	   // In server range ?
	   if ( (ntohl(req_align)>=ntohl(server_config.start)) && (ntohl(req_align)<=ntohl(server_config.end)) )
	   {
            // and its not already taken/offered
	    if ( !(lease = find_lease_by_yiaddr(req_align)) || lease_expired(lease) )
     	    {
	     packet.yiaddr = req_align; // Giving this adress
	     LOG(LOG_INFO,"sendOffer","You asked it, you got it : '%s'",DHCPNumber2String(packet.yiaddr,ip));
	     goto ready;
	    }
	    else LOG(LOG_WARNING,"sendOffer","Wanted address '%s' is taken or offered",DHCPNumber2String(req_align,ip));
	   }
	   else LOG(LOG_WARNING,"sendOffer","Wanted address '%s' is out of range",DHCPNumber2String(req_align,ip));
	  } // End of if(req)
          else LOG(LOG_INFO,"sendOffer","No Wanted address");

	  // Otherwise find an adress
	  LOG(LOG_INFO,"sendOffer","Seeking a new address");
	  packet.yiaddr = find_address(0);

	  // try for an expired lease
	  if (!packet.yiaddr) packet.yiaddr = find_address(1);
	  LOG(LOG_INFO,"sendOffer","Find address '%s'",DHCPNumber2String(packet.yiaddr,ip));
	}
ready:
	if(!packet.yiaddr)
	{
		LOG(LOG_WARNING,"sendOffer","No IP address to give -- OFFER abandoned");
		return (-1);
	}
	// Memorize lease
	if (!add_lease(packet.chaddr, packet.yiaddr, server_config.offer_time,sizeof(packet.chaddr)))
	{
		LOG(LOG_WARNING,"sendOffer","Lease pool is full -- OFFER abandoned");
		return (-1);
	}

	if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME)))
        {
		LOG(LOG_INFO,"sendOffer","Lease time asked is :%d",*lease_time);
		memcpy(&lease_time_align, lease_time,sizeof(lease_time_align));
		lease_time_align = ntohl(lease_time_align);
		if (lease_time_align > server_config.lease)
			lease_time_align = server_config.lease;
	}
	// Make sure we aren't just using the lease time from the previous offer
	if (lease_time_align < server_config.min_lease)
		lease_time_align = server_config.min_lease;
	LOG(LOG_INFO,"sendOffer","Lease time set is:%d",lease_time_align);
	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

	add_server_options(&packet);

	add_bootp_options(&packet);
	LOG(LOG_INFO,"sendOffer","BOOT Options added");

	LOG(LOG_INFO,"sendOffer","Sending OFFER of %s for %d s", DHCPNumber2String(packet.yiaddr,ip),lease_time_align);
	return send_packet(&packet, 0);
}

//------------------------------------------------------------------------------------------------------
// Sending a NAK message
// Returns number of bytes sent or -1
//------------------------------------------------------------------------------------------------------
int sendNAK(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;		// Packet to send
	DEBUG("sendNAK","Called");
	init_packet(&packet, oldpacket, DHCPNAK);
	return (send_packet(&packet, 1));
}

//------------------------------------------------------------------------------------------------------
// Sending a ACK message to yiaddr
// Returns 0 if allright , -1 if failed
//------------------------------------------------------------------------------------------------------
int sendACK(struct dhcpMessage *oldpacket, u_int32_t yiaddr)
{
	struct dhcpMessage packet;		// Packet to send
	unsigned char *lease_time=NULL;		// Pointer on LEASE_TIME
	u_int32_t lease_time_align = server_config.lease; // LEASE_TIME
#ifdef LOGACTIVE
	char ip[IPSTRING];		// IP adress conversion
#endif

	init_packet(&packet, oldpacket, DHCPACK);	// Prepare a simple ACK packet
	packet.yiaddr = yiaddr;				// set target

	DEBUG("sendACK","Called");
	if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME)))
	{
		LOG(LOG_INFO,"sendACK","Lease time asked was :%d",*lease_time);
		memcpy(&lease_time_align, lease_time,sizeof(lease_time_align));
		lease_time_align = ntohl(lease_time_align);
		// Check boundaries
		if (lease_time_align > server_config.lease)
			lease_time_align = server_config.lease;
		else if (lease_time_align < server_config.min_lease)
			lease_time_align = server_config.min_lease;
	}

	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

	add_server_options(&packet); // Add Server options

	add_bootp_options(&packet); // Add bootp options

	LOG(LOG_INFO,"sendACK","Sending ACK to %s",DHCPNumber2String(packet.yiaddr,ip));

	if (send_packet(&packet, 0) < 0)
	 return (-1);

	add_lease(packet.chaddr, packet.yiaddr, lease_time_align,sizeof(packet.chaddr)); // Memorize it

	return (0);
}

//------------------------------------------------------------------------------------------------------
// Sending information message
// Returns number of bytes sent or -1
//------------------------------------------------------------------------------------------------------
int send_inform(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;		// Packet to send

	DEBUG("send_inform","Called");
	init_packet(&packet, oldpacket, DHCPACK); // Prepare a simple ACK packet

	add_server_options(&packet); // Add server options

	add_bootp_options(&packet); 	// Add bootp options

	return (send_packet(&packet,0));
}


