/*
 * socket.c -- DHCP server client/server socket creation
 *
 * udhcp client/server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <network.h>
#include <net/ethernet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <errno.h>
#include "ecos.h"
#include <net/dhcpelios/debug.h>

//-----------------------------------------------------------------------------------------
// Readindex (ifindex) ,IP address (addr) and MAC adress (arp) of a network interface (interface)
// @Return 0 on success, -1 if failed
//-----------------------------------------------------------------------------------------
char read_interface(char *interface, int *ifindex, u_int32_t *addr, unsigned char *arp)
{
	int fd=0;
	struct ifreq ifr;
	struct sockaddr_in *our_ip=NULL;

#ifdef LOGACTIVE
 char arpad[MACSTRING];
#endif

	DEBUG("read_interface","Called with interface :'%s'",interface);
	memset(&ifr, 0, sizeof(struct ifreq));

	if((fd = socket(AF_INET, SOCK_DGRAM, 0)) >= 0)
	{
		ifr.ifr_addr.sa_family = AF_INET;
		strncpy(ifr.ifr_name, interface,IFNAMSIZ);

		if (addr)
		{
			// Get interface address
			if (ioctl(fd, SIOCGIFADDR, &ifr) == 0)
			{
				our_ip = (struct sockaddr_in *) &ifr.ifr_addr;
				*addr = our_ip->sin_addr.s_addr;
				LOG(LOG_INFO, "read_interface","Our IP on '%s' is '%s'", ifr.ifr_name, inet_ntoa(our_ip->sin_addr));
			}
			else
			{
				LOG(LOG_ERR, "read_interface"," SIOCGIFADDR failed, is the interface '%s' up and configured ?: '%s'",interface,strerror(errno));
				close(fd);
				return (-1);
			}
		}
		// Interface index
		*ifindex = if_nametoindex(interface);
		LOG(LOG_INFO,"read_interface", "Adapter index is %d",*ifindex);

		// Get Hardware adress (MAC)
		if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0)
		{
		 memcpy(arp, ifr.ifr_hwaddr.sa_data,ETHER_ADDR_LEN);
		 LOG(LOG_INFO,"read_interface", "Adapter hardware address '%s'",DHCPether_print(arp,arpad,MACSTRING));
		}
		else
		{
		 LOG(LOG_ERR,"read_interface", "SIOCGIFHWADDR failed!: '%s'", strerror(errno));
	 	 close(fd);
		 return (-1);
		}
  	 close(fd);
	} else
	{
	 LOG(LOG_ERR, "read_interface","Getting interface informations failed!: '%s'", strerror(errno));
	 return (-1);
	}
	return (0);
}

//-----------------------------------------------------------------------------------------
// Read mask address (addr) of a network interface (interface)
// @Return 0 on success, -1 if failed
//-----------------------------------------------------------------------------------------
char read_interface_mask(char *interface, u_int32_t *addr)
{
	int fd=0;
	struct ifreq ifr;
	struct sockaddr_in *our_mask=NULL;

	DEBUG("read_interface_mask","Called with interface :'%s'",interface);
	memset(&ifr, 0, sizeof(struct ifreq));

	if((fd = socket(AF_INET, SOCK_DGRAM, 0)) >= 0)
	{
		ifr.ifr_addr.sa_family = AF_INET;
		strncpy(ifr.ifr_name, interface,IFNAMSIZ);

		if (addr)
		{
			// Get interface address
			if (ioctl(fd, SIOCGIFNETMASK, &ifr) == 0)
			{
				our_mask = (struct sockaddr_in *) &ifr.ifr_addr;
				*addr = our_mask->sin_addr.s_addr;
				LOG(LOG_INFO, "read_interface","Our mask on '%s' is '%s'", ifr.ifr_name, inet_ntoa(our_mask->sin_addr));
			}
			else
			{
				LOG(LOG_ERR, "read_interface"," SIOCGIFNETMASK failed, is the interface '%s' up and configured ?: '%s'",interface,strerror(errno));
				close(fd);
				return (-1);
			}
		}
  	 close(fd);
	} else
	{
	 LOG(LOG_ERR, "read_interface_mask","Getting interface informations failed!: '%s'", strerror(errno));
	 return (-1);
	}
	return (0);
}

//-----------------------------------------------------------------------------------------
// Create a listening UDP broadcast socket on selected ip port on interface named inf
// Return -1 on error, or socket file descriptor number
//-----------------------------------------------------------------------------------------
int listen_socket(unsigned int ip, int port, char *inf)
{
	int fd=0;
	struct sockaddr_in addr;
	int n = 1;
#ifdef DEBUGGING
	char ips[IPSTRING];
#endif

	DEBUG("listen_socket","Called with '%s':%d inf='%s'",DHCPNumber2String(ip,ips), port, inf);
	if ((fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	{
	 LOG(LOG_ERR,"listen_socket", "Socket call failed: '%s'", strerror(errno));
	 return (-1);
	}
	LOG(LOG_INFO, "listen_socket","Listen socket is %d", fd);

	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &n, sizeof(n)) == -1)
	{
	 close(fd);
	 LOG(LOG_ERR,"listen_socket", "Setsockopt for REUSEADDR failed: '%s'",strerror(errno));
	 return (-1);
	}
 	LOG(LOG_INFO,"listen_socket", "Socket will work for BROADCAST");
	if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (char *) &n, sizeof(n)) == -1)
	{
	 close(fd);
	 LOG(LOG_ERR,"listen_socket", "Setsockopt for BROADCAST failed :'%s'",strerror(errno));
	 return (-1);
	}

	// Adresse Source
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = ip;
	addr.sin_len=sizeof(addr);
	if (bind(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr)) == -1)
	{
		LOG(LOG_ERR,"listen_socket", "Bind failed :'%s'",strerror(errno));
		close(fd);
		return (-1);
	}
	return (fd);
}
